/*
    Digg Sidebar - Shows Digg stories in the Firefox sidebar in real time.
    Copyright (C) 2008  Abhinav Sarkar <abhinav dot sarkar at gmail dot com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

DiggSidebar = {
    prefs: Application.extensions.get("diggsidebar@abhinavsarkar.net").prefs,
    storyListRefreshEventTopic: "extension-diggsidebar-abhinavsarkar-net-storylist-refresh",
    endpoint: {
        container: null,
        topic: null,
        category: "all"
    },
    updateInterval: 10,
    updateIntervalDecay: 0,
    shownStoriesCount: 30,
    categories: ['All', 'Popular', 'Upcoming', 'Hot', 'Top'],
    timerId: null,
    storyIds: [],
    playing: true,
    stories: [],

    fetchData: function(url, handler) {
        DiggSidebar.diggIndicator.style.display = '';

        var request = new XMLHttpRequest();
        request.onerror = DiggSidebar.errorHandler;
        request.onload = handler;
        request.open("GET", url, true);
        request.send(null);
    },

    errorHandler: function(e) {
        window.alert("Error in accessing data from Digg");
    },

    populateMenu: function(e) {
        var XHR = e.target;
        var data = DiggSidebar.Utils.decodeJson(XHR.responseText);
        var topics = data.topics;
        var containers = new Array();

        while (DiggSidebar.topicPopup.firstChild) {
                DiggSidebar.topicPopup.removeChild(DiggSidebar.topicPopup.firstChild);
        }

        topics.forEach(function (topic) {
            var menu = document.createElement('menu');
            menu.setAttribute('id', topic.short_name + 'Menu');
            menu.setAttribute('label', topic.name);
            menu.setAttribute('accesskey', topic.name.charAt(0));

            var menupopup = document.createElement('menupopup');
            menupopup.setAttribute('id', topic.short_name + 'Popup');

            DiggSidebar.categories.forEach(
                function (label) {
                    var menuitem = document.createElement('menuitem');
                    menuitem.setAttribute('label', label);
                    menuitem.setAttribute('value', '/topic/' + topic.short_name + '/' + label.toLowerCase());
                    menuitem.setAttribute('accesskey', label.charAt(0));
                    menuitem.setAttribute('oncommand', "DiggSidebar.setEndPoint(this.value)");
                    menupopup.appendChild(menuitem);
                }
            );
            menu.appendChild(menupopup);
            DiggSidebar.topicPopup.appendChild(menu);

            containers.push(topic.container);
        });

        while (DiggSidebar.containerPopup.firstChild) {
                DiggSidebar.containerPopup.removeChild(DiggSidebar.containerPopup.firstChild);
        }

        var filteredContainers = {name: [], short_name: []};
        containers.forEach(function (container) {
            var name = container.name;
            var short_name = container.short_name;
            if (filteredContainers.name.indexOf(name) == -1) {
                filteredContainers.name.push(name);
                filteredContainers.short_name.push(short_name);
            }
        });

        for (var i=0; i<filteredContainers.name.length; i++) {
            var menu = document.createElement('menu');
            menu.setAttribute('id', filteredContainers.short_name[i] + 'Menu');
            menu.setAttribute('label', filteredContainers.name[i]);
            menu.setAttribute('accesskey', filteredContainers.name[i].charAt(0));

            var menupopup = document.createElement('menupopup');
            menupopup.setAttribute('id', filteredContainers.short_name[i] + 'Popup');

            DiggSidebar.categories.forEach(
                function (label) {
                    var menuitem = document.createElement('menuitem');
                    menuitem.setAttribute('label', label);
                    menuitem.setAttribute('value', '/container/' +
                        filteredContainers.short_name[i] + '/' + label.toLowerCase());
                    menuitem.setAttribute('accesskey', label.charAt(0));
                    menuitem.setAttribute('oncommand', "DiggSidebar.setEndPoint(this.value)");
                    menupopup.appendChild(menuitem);
                }
            );

            menu.appendChild(menupopup);
            DiggSidebar.containerPopup.appendChild(menu);
        }
        DiggSidebar.diggIndicator.style.display = 'none';
    },

    populateStoryList: function() {
        while (DiggSidebar.storyList.firstChild) {
            DiggSidebar.storyList.removeChild(DiggSidebar.storyList.firstChild);
        }

        var newStoryIds = new Array();
        Application.console.log(DiggSidebar.stories.length);

        var jp = new JPath(DiggSidebar.stories);
        var filtedStories = DiggSidebar.stories;
        if (DiggSidebar.endpoint.topic)
            filtedStories = jp.$(function(story){
                return story.$("topic/short_name").json == DiggSidebar.endpoint.topic;
            }).json;
        if (DiggSidebar.endpoint.container)
            filtedStories = jp.$(function(story){
                return story.$("container/short_name").json == DiggSidebar.endpoint.container;
            }).json;

        filtedStories.forEach(function (story, index) {
            if (index < DiggSidebar.shownStoriesCount) {
                var listitem = DiggSidebar.storyList.appendChild(document.createElement('richlistitem'));
                listitem.id = "story_" + story.id;
                listitem.setAttribute('title', story.title);
                if (DiggSidebar.storyIds.indexOf(story.id) != -1) listitem.new = false;
            }
            if (DiggSidebar.storyIds.indexOf(story.id) == -1) newStoryIds.push(story.id);
        });
        DiggSidebar.storyIds = DiggSidebar.storyIds.concat(newStoryIds);
        DiggSidebar.setUpdateInterval(DiggSidebar.stories.slice(-5));
        DiggSidebar.diggIndicator.style.display = 'none';
    },

    populateDescription: function(e) {
        var XHR = e.target;
        //Application.console.log(XHR.responseText);
        var data = DiggSidebar.Utils.decodeJson(XHR.responseText);
        var story = data.stories[0];

        var now = new Date();
        if (story.promote_date != null)
            var then = new Date(story.promote_date*1000);
        else
            var then = new Date(story.submit_date*1000);
        var diff = Math.max(now - then, 0)

        var hr = Math.floor(diff/(1000*3600));
        var min = Math.floor(diff/(1000*60)) - 60*hr;

        var listitems = document.getElementsByTagName('richlistitem');
        for (var i=0; i<listitems.length; i++) {
                listitems[i].hideDesc();
        }

        var li = document.getElementById('story_'+ story.id)

        var relativeTime = ((hr > 0) && (min > 0)) ?
            (hr + " hr " + min + " mins ago") :
            ((hr == 0) && (min > 0)) ?
            (min + " mins ago") :
            ((hr == 0) && (min == 0)) ?
            "just now" : "";
        li.setAttribute('date', relativeTime);
        li.setAttribute('status', story.status);
        li.setAttribute('container', story.container.name);
        li.setAttribute('topic', story.topic.name);
        li.setAttribute('username', story.user.name);
        li.setAttribute('userlink', "http://digg.com/users/" + story.user.name);
        li.setAttribute('diggs', story.diggs);
        li.setAttribute('comments', story.comments);
        li.setAttribute('desc', story.description);
        li.setAttribute('link', story.link);
        li.setAttribute('href',  story.href);

        li.read = true;
        li.new = false;
        li.showDesc()

        DiggSidebar.diggIndicator.style.display = 'none';
    },

    getStories: function() {
        var ep = DiggSidebar.prefs.get("endpoint").value || '';
        DiggSidebar.diggEndPoint.value = "digg" + ep;

        DiggSidebar.fetchData("http://services.digg.com/stories" + ep.replace(/\/all/g, '') +
            "?count=30" + "&type=json" +
            "&appkey=" + encodeURIComponent("http://diggsidebar.googlepages.com"),
            function(e) {
                var newStories = DiggSidebar.Utils.decodeJson(e.target.responseText).stories.filter(
                    function(story) {
                        return !(DiggSidebar.storyIds.indexOf(story.id) != -1);
                    });
                newStories.forEach(function(story) {
                    story.category = DiggSidebar.endpoint.category;
                    story.read = false;
                });
                DiggSidebar.stories = newStories.concat(DiggSidebar.stories);
                var observerService = Components.classes["@mozilla.org/observer-service;1"]
                    .getService(Components.interfaces.nsIObserverService);
                observerService.notifyObservers(null, DiggSidebar.storyListRefreshEventTopic, null);
            });
    },

    getDescription: function(storyId) {
        if (storyId == null) return;
        DiggSidebar.fetchData("http://services.digg.com/story/" + storyId + "?type=json" +
            "&appkey=" + encodeURIComponent("http://diggsidebar.googlepages.com"),
            DiggSidebar.populateDescription);
    },

    createMenu: function() {
        DiggSidebar.fetchData("http://services.digg.com/topics" + "?type=json" +
            "&appkey=" + encodeURIComponent("http://diggsidebar.googlepages.com"),
            DiggSidebar.populateMenu);
    },

    setUpdateInterval: function(stories) {
        //Adaptive update interval code START
        if (stories.length == 0) {
            window.clearTimeout(DiggSidebar.timerId);
            var timeout = Math.round(DiggSidebar.updateInterval*(Math.pow(1.5, DiggSidebar.updateIntervalDecay)));
            DiggSidebar.timerId = window.setTimeout(DiggSidebar.getStories, timeout);
            return;
        }
        var sum = 0;
        var weights = [4, 3, 2, 1]
        var date = (stories[0].promote_date != null) ? 'promote_date' : 'submit_date';
        stories.sort(function (a, b) {
            if (a[date] > b[date]) return -1;
            else if (a[date] < b[date]) return 1;
            return 0;
        });
        for (var i=0; i<Math.min(4, stories.length-1); i++) {
            var diff = stories[i][date] - stories[i+1][date];
            sum += weights[i]*diff;
        }
        var updateInterval = Math.round(
            sum*1000/weights
            .splice(0,Math.min(4, stories.length-1))
            .reduce(function(a, b){return a + b;})
        );
        //Application.console.log(updateInterval);

        if (updateInterval > 0) {
            var previousUpdateInterval = DiggSidebar.updateInterval;
            DiggSidebar.updateInterval = updateInterval;

            var updateIntervalDecay = DiggSidebar.updateIntervalDecay;

            if (previousUpdateInterval == updateInterval)
                DiggSidebar.updateIntervalDecay = updateIntervalDecay + 1;
            else
                DiggSidebar.updateIntervalDecay = 0;
        }
        window.clearTimeout(DiggSidebar.timerId);

        var timeout = Math.round(DiggSidebar.updateInterval*(Math.pow(1.5, DiggSidebar.updateIntervalDecay)));
        DiggSidebar.timerId = window.setTimeout(DiggSidebar.getStories, timeout);
        //Adaptive update interval code END
    },

    setEndPoint: function(ep) {
        DiggSidebar.prefs.get("endpoint").value = ep;
        DiggSidebar.endpoint = DiggSidebar.getEndpointParts(ep);
        window.clearTimeout(DiggSidebar.timerId);
        DiggSidebar.getStories();
    },

    getEndpointParts: function(ep) {
        var tmp = {
            topic: null,
            container: null,
            category: "all"
        };
        parts = ep.split("/");
        if (parts.length == 2)
            tmp.category = parts[1];
        else if (parts.length == 4) {
            tmp[parts[1]] = parts[2];
            tmp.category = parts[3];
        }
        return tmp;
    },

    togglePlayPause: function() {
        if (DiggSidebar.playing) {
            window.clearTimeout(DiggSidebar.timerId);
            DiggSidebar.diggPlayPause.image = "chrome://diggsidebar/content/image/Play.png";
            DiggSidebar.diggPlayPause.setAttribute("tooltiptext", "Click to Start autoupdate");
            DiggSidebar.playing = false;
        } else {
            var timeout = Math.round(DiggSidebar.updateInterval*(Math.pow(1.5, DiggSidebar.updateIntervalDecay)));
            DiggSidebar.timerId = window.setTimeout(DiggSidebar.getStories, timeout);
            DiggSidebar.diggPlayPause.image = "chrome://diggsidebar/content/image/Pause.png";
            DiggSidebar.diggPlayPause.setAttribute("tooltiptext", "Click to Pause autoupdate");
            DiggSidebar.playing = true;
        }
    },

    initialize: function(){
        DiggSidebar.endpoint = DiggSidebar.getEndpointParts(DiggSidebar.prefs.get("endpoint").value);
        var $ = function(id) {return document.getElementById(id)};
        DiggSidebar.diggIndicator = $('diggIndicator');
        DiggSidebar.topicPopup = $('topicPopup');
        DiggSidebar.containerPopup = $('containerPopup');
        DiggSidebar.storyList = $('storyList');
        DiggSidebar.diggEndPoint = $('diggEndPoint');
        DiggSidebar.diggPlayPause = $('diggPlayPause');

        DiggSidebar.storyListObserver = new StoryListObserver();
        DiggSidebar.createMenu();
        DiggSidebar.getStories();
    },

    destroy: function(){
        DiggSidebar.storyListObserver.unregister();
    }
};

DiggSidebar.Utils = {
    url: function(spec) {
        var ios = Components.classes["@mozilla.org/network/io-service;1"]
                    .getService(Components.interfaces.nsIIOService);
        return ios.newURI(spec, null, null);
    },

    openUrlInTab: function(url) {
        Application.activeWindow.open(DiggSidebar.Utils.url(url));
    },

    decodeJson: function(string) {
        var json = Components.classes["@mozilla.org/dom/json;1"]
            .createInstance(Components.interfaces.nsIJSON);
        return json.decode(string);
    }
}

var StoryListObserver = function() {
   this.registered = false;
   this.register();
}

StoryListObserver.prototype = {
    observe: function(subject, topic, data) {
        if (topic == window.DiggSidebar.storyListRefreshEventTopic) {
            window.DiggSidebar.populateStoryList();
        }
    },

    register: function() {
        if (this.registered == false) {
            var observerService = Components.classes["@mozilla.org/observer-service;1"]
                .getService(Components.interfaces.nsIObserverService);
            observerService.addObserver(this, DiggSidebar.storyListRefreshEventTopic, false);
            this.registered = true;
        }
    },

    unregister: function() {
        if (this.registered == true) {
            var observerService = Components.classes["@mozilla.org/observer-service;1"]
                .getService(Components.interfaces.nsIObserverService);
            observerService.removeObserver(this, DiggSidebar.storyListRefreshEventTopic);
            this.registered = false;
        }
    }
}

/*TODO
 preferences
 toolbar button
 refresh button
 */
/*DONE
 v0.2
 added kdb navigation
 Added progressmeter
 added error notification
 added accesskeys
 added new story notification

 v0.2.1
 adaptive polling interval
 promote_date
 pause/play

 v0.5
 new ui
*/
